{ Die RCIT (Real Color Index - Technologie):                  }
{ ------------------------------------------                  }
{                                                             }
{ Die Unit verwendet einen Algorithmus, der es mglich macht, }
{ die Farben eines Bitmaps ber die direkte Farbnummer anzu-  }
{ sprechen. Vorher gab es folgendes Problem:                  }
{ ----------------------------------------------------------- }
{        Farbe 1   --> r: 176, g: 176, b: 176                 }
{        Farbe 247 --> r: 176, g: 176, b: 176                 }
{ Wurde nun mit Farbe 247 gezeichnet, speicherte Windows im   }
{ Speicherbitmap die Farbnummer 1 ab, da er aufgrund des      }
{ RGB(176, 176, 176)-Makros die erste Farbe in der Bitmap-    }
{ farbpalette verwendet, welche im Handle enthalten ist.      }
{                                                             }
{ Nun die neue Methode:                                       }
{ ----------------------------------------------------------- }
{ Bei der Erzeugung eines neuen Bitmaps verwende ich einfach  }
{ eine simple Farbpalette, welche von Schwarz in komplett rot }
{ bergeht:                                                   }
{        Farbe 0   --> r: 0, g: 0, b: 0                       }
{        Farbe 1   --> r: 1, g: 0, b: 0                       }
{        Farbe 2   --> r: 2, g: 0, b: 0                       }
{        Farbe 3   --> r: 3, g: 0, b: 0                       }
{        Farbe 4   --> r: 4, g: 0, b: 0                       }
{        Farbe 5   --> r: 5, g: 0, b: 0                       }
{           .                                                 }
{           .                                                 }
{           .                                                 }
{        Farbe 255 --> r: 255, g: 0, b: 0                     }
{ Desweiteren verwende ich eine zweite Farbpalette, welche    }
{ ich ScreenColorTable nannte. Will ich nun mit der Farbe 247 }
{ im Bitmap zeichnen, so setze ich einfach die entsprechende  }
{ Farbe mittels des RGB(r, g, b)-Makros. Will ich z.B. einen  }
{ Punkt im Bitmap mit der Farbnummer 247 zeichnen, so gengt  }
{ zur Farbfestlegung folgender Aufruf: RGB(247, 0, 0).        }
{ Daraus schlufolgere ich, da der im RGB-Makro festgelegte  }
{ Rot-Anteil der Farbnummer entspricht. Der Grn- bzw. Blau-  }
{ anteil mu immer 0 sein.                                    }
{ Beim Anzeigen dieses Bitmaps wird nun die ScreenColorTable  }
{ bentigt. In dieser stehen die richtigen Farben. Die ver-   }
{ wendete API-Funktion (StretchDIBits) von Windows inter-     }
{ pretiert nun die im Bitmap stehenden Farbnummern als eine   }
{ Farbe der ScreenColorTable und zeigt somit das Bitmap in    }
{ der richtigen Form an. Ein Hinweis sei jedoch gegeben:      }
{ Sollte mittels BitBlt, CopyRect oder anderer Funktionen von }
{ einem Delphi-Farbbitmap (TImage, TBitmap, ...) etwas        }
{ kopiert werden, sollte mittels der Methode SetDIBColorTable }
{ die richtige Farbpalette des Bitmaps gesetzt werden, da     }
{ Windows sonst versucht die Farben an die Rotfarbenpalette   }
{ anzupassen. Am Ende aber nicht vergessen diese mittels des  }
{ Aufrufes SetDIBColorTable(RedColorTable, 0, 255) wieder zu  }
{ restaurieren.                                               }


unit RCIBitmap;

interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, Menus;

type
  { Farbpalette mit 256 Eintrgen }
  TRCIEntries = array[0..255] of TRGBQuad;

  { Pointer auf Paletteneintrge als Typ }
  PRCIEntries = ^TRCIEntries;

  { ein Farbpaletteneintrag }
  RGBColors   = record
                  rgbRed  : BYTE;
                  rgbGreen: BYTE;
                  rgbBlue : BYTE;
                end;

  {              PCX-InfoHeader                    }
  { ---------------------------------------------- }
  { MAN BEACHTE:                                   }
  {   - Breite des PCX-Bildes MU immer durch zwei }
  {     teilbar sein!!!!!!!!!!!!!!!! Wenn nicht,   }
  {     dann auf nchsten Teiler aufrunden!        }
  PCXHeader   = record
                  Id,                 { Identifikation des Herstellers }
                  Ver,                { Versionsnummer (5=Version 3.0) }
                  Encod,              { Kodierung (1=Runlength Kodierung) }
                  BitProPixel: byte;  { 1=16 Farben;8=256 Farben }
                  xmin, ymin,         { Eckkoordinaten des Originalbildes }
                  xmax,  ymax,
                  hres,
                  vres: word;         { Auflsung }
                  Farbe: array[0..15] of RGBColors;
                                      { 16 Farben Palette }
                  Fill: byte;         { reserviert }
                  Planes: byte;       { Anzahl Farbebenen }
                  ByteProZeile: word; { ... pro Plane }
                  PalInfo: word;      { Palettentyp (1=Farbe;2=SW) }
                  Fill3: array[0..57] of byte;
                                      { reserviert }
                end;

  { MAN BEACHTE BEI BITMAPS:                       }
  {   - Breite des BMP-Bildes MU immer durch vier }
  {     teilbar sein!!!!!!!!!!!!!!!! Wenn nicht,   }
  {     dann auf nchsten Teiler aufrunden und am  }
  {     Ende wieder abziehen (nach laden, ...)!!!! }
  TRCIBitmap  = class
                  public
                    Canvas       : TCanvas;
                    OrigDIBBits  : POINTER;
                    NewDIBBits   : POINTER;
                    DIB          : HBitmap;
                    DIBHeader    : TBitmapFileHeader;
                    DIBInfoHeader: ^TBitmapInfo;
                    LastFormat   : Integer;
                    ClipboardData: TMemoryStream;
                    constructor Create;
                    destructor Destroy;
                    procedure Draw(ACanvas: TCanvas;
                                   Left, Top, DrawMode: Longint);
                    procedure NewPicture(Width, Height: word;
                                         Background: BYTE;
                                         ColTable: TRCIEntries);
                    procedure SetScreenColorTable(Table: TRCIEntries;
                                                  FromClr, ToClr: BYTE);
                    procedure SetDIBColorTable(Table: TRCIEntries;
                                               FromClr, ToClr: Integer);
                    procedure CopyToClipboard(Pos: TRect);
                    function PastePicture: BOOLEAN;
                    function LoadFromFile(const Filename: string): boolean;
                    function SaveToFile(const Filename: string;
                                        Format: Integer): boolean;
                  private
                    OrigBMP: Integer;
                    MehrPix: Integer;
                    procedure CheckMem;
                    procedure FillHeader(Width, Height: word);
                    procedure LoadBMPFromFile(const Filename: string);
                    function LoadPCXFromFile(const Filename: string): BOOLEAN;
                    procedure SaveBMPToFile(const Filename: string);
                    procedure SavePCXToFile(const Filename: string);
                end;

const
  { Bild als BMP abspeichern }
  BMPFormat        = 1;
  { Bild als PCX abspeichern }
  PCXFormat        = 2;
  { Soll Bildformat beim abspeichern automatisch erkannt werden }
  AutoDetectFormat = 3;
  { Kennung fr mein Format (in Clipboard) }
  ClipboardDetect  = $C0;

var
  a            : BYTE;
  RedColorTable: TRCIEntries; { rotfarb-Palette }

implementation

constructor TRCIBitmap.Create;
begin
  { vorherige Create-Routine aufrufen }
  inherited Create;
  { DIB-Handle = 0 }
  DIB:= 0;
  { Canvas-Objekt erzeugen }
  Canvas:= TCanvas.Create;
  { Objekt einen Handle zuweisen }
  Canvas.Handle:= CreateCompatibleDC(GetDC(0));
  { Speicher fr DIB-Header reservieren }
  GetMem(DIBInfoHeader, SizeOf(TBitmapInfo)+ 1020);
  { Originalbitmap speichern }
  OrigBMP:= GetCurrentObject(Canvas.Handle, OBJ_BITMAP);
  { Pixel, die zuviel gespeichert wurden auf null setzen }
  MehrPix:= 0;
  { Letztes Format war BMP }
  LastFormat:= BMPFormat;
end;

destructor TRCIBitmap.Destroy;
var
  test : BOOLEAN; { wurde Handle ordnungsgem zerstrt? }
  OldDC: HDC;     { altes DC-Handle }
begin
  { altes Handle merken }
  OldDC:= Canvas.Handle;
  { DIB-Objekt (falls belegt zerstren }
  CheckMem;
  { ... und zerstren }
  Canvas.Destroy;
  Canvas:= nil;
  { nur lschen, wenn belegt }
  if OldDC <> 0 then
  begin
    { Canvashandle zerstren }
    test:= DeleteDC(OldDC);
    { konnte ordentlich zerstrt werden? }
    if test = FALSE then
      { Nein!! ... dann Fehlermeldung }
      Application.MessageBox('Couldn`t destroy DC-Handle!',
                             'Error', MB_OK);
  end;
  { ist DIBInfoHeader im RAM reserviert }
  if DIBInfoHeader <> nil then
  begin
    { ... dann zerstrt }
    FreeMem(DIBInfoHeader, SizeOf(TBitmapInfo)+ 1020);
    { Pointer null setzen }
    DIBInfoHeader:= nil;
  end;
  { vorherige Destroy-Routine aufrufen }
  inherited Destroy;
end;

procedure TRCIBitmap.CheckMem;
begin
  { DIB-Handle da }
  if DIB <> 0 then
  begin
    { Originalbitmap selektieren }
    SelectObject(Canvas.Handle, OrigBMP);
    { ... dann lschen }
    DeleteObject(DIB);
    { DIB-Handle null setzen }
    DIB:= 0;
  end;
end;

procedure TRCIBitmap.Draw(ACanvas: TCanvas;
                          Left, Top, DrawMode: Longint);
begin
  { ist DIB-Handle belegt? }
  if DIB <> 0 then
    with DIBInfoHeader^.bmiHeader do
      { ... dann DIB zeichnen }
      StretchDIBits(ACanvas.Handle, Left, Top, biWidth, biHeight,
                    0, 0, biWidth, biHeight, NewDIBBits,
                    DIBInfoHeader^, DIB_RGB_COLORS, DrawMode);
end;

procedure TRCIBitmap.SetScreenColorTable(Table: TRCIEntries;
                                         FromClr, ToClr: BYTE);
var
  a: integer; { simpler Laufindex }
begin
  { Anzeigefarbpalette setzen }
  for a:= FromClr to ToClr do
    DIBInfoHeader.bmiColors[a]:= Table[a];
end;

procedure TRCIBitmap.SetDIBColorTable(Table: TRCIEntries;
                                      FromClr, ToClr: Integer);
begin
  { interne DIB-Farbpalette setzen }
  Windows.SetDIBColortable(Canvas.Handle, FromClr, ToClr- FromClr+ 1,
                           Table[0]);
end;

procedure TRCIBitmap.FillHeader(Width, Height: word);
begin
  { ist ein DIB-Info-Header angelegt? }
  if DIBInfoHeader = nil then
    { Nein!! ... dann einen erzeugen }
    GetMem(DIBInfoHeader, SizeOf(TBitmapInfo)+ 1020);
  { DIB-Header mit Werten belegen }
  with DIBHeader do
  begin
    { ... als erstes den Typ }
    bfType:= 19778;
    { ... nun Dateigre }
    bfSize:= SizeOf(TBitmapFileHeader)+ SizeOf(TBitmapInfo)+
               Width* Height+ 1020;
    { ... 2 Reserviertbytes }
    bfReserved1:= 0;
    bfReserved2:= 0;
    { ... und als letztes Start der Bilddaten }
    bfOffBits:= SizeOf(TBitmapFileHeader)+ SizeOf(TBitmapInfo)+ 1020;
  end;
  { DIBInfoHeader belegen }
  with DIBInfoHeader^.bmiHeader do
  begin
    { ... Gre des InfoHeaders }
    biSize:= SizeOf(DIBInfoHeader^.bmiHeader);
    { ... Breite }
    biWidth:= Width;
    { ... Hhe }
    biHeight:= Height;
    { ... Anzahl Planes }
    biPlanes:= 1;
    { ... Anzahl Bits pro Pixel angeben (immer 256 Farben) }
    biBitCount:= 8;
    { ... Kompression }
    biCompression:= BI_RGB;
    { ... Gre des Bildes berechnen }
    biSizeImage:= Width* Height;
    { ... Pixel pro Meter (0, wird nicht bentigt) }
    biXPelsPerMeter:= 0;
    biYPelsPerMeter:= 0;
    { ... verwendete Farben }
    biClrUsed:= 0;
    { ... wichtige Farben (0, da alle Farben wichtig sind) }
    biClrImportant:= 0;
  end;
end;

procedure TRCIBitmap.NewPicture(Width, Height: word;
                                Background: BYTE;
                                ColTable: TRCIEntries);
var
  a         : integer;      { simpler Laufindex }
  InfoHeader: ^TBitmapInfo; { InfoHeader mit RedColorTable }
begin
  { temporren Bitmapinfo-Header erzeugen }
  GetMem(InfoHeader, SizeOf(TBitmapInfo)+ 1020);
  { Ressourcen checken }
  CheckMem;
  { DIB-Header mit Werten belegen }
  FillHeader(Width, Height);
  { temporren BitmapInfoHeader mit gleichen Daten belegen }
  InfoHeader^:= DIBInfoHeader^;
  { Farbpalette mit Farben belegen }
  for a:= 0 to 255 do
  begin
    { aktuelle Farbpalette setzen }
    DIBInfoHeader^.bmiColors[a]:= ColTable[a];
    { interne Palette mit Daten fllen }
    InfoHeader^.bmiColors[a]:= RedColorTable[a];
  end;
  { Bitmap erzeugen }
  DIB:= CreateDIBSection(Canvas.Handle, InfoHeader^,
                         DIB_RGB_COLORS, NewDIBBits, nil, 0);
  { Objekt selektieren }
  SelectObject(Canvas.Handle, DIB);
  { ... und Hintergrundfarbe zeichnen }
  Canvas.Pen.Color:= RGB(Background, 0, 0);
  { ... und Hintergrundfarbe zeichnen }
  Canvas.Brush.Color:= Canvas.Pen.Color;
  { ... und zeichnen auslsen }
  Canvas.FillRect(Rect(0, 0, Width, Height));
  { temporren Bitmapinfo-Header wieder lschen }
  FreeMem(InfoHeader, SizeOf(TBitmapInfo)+ 1020);
  { InfoHeaderangaben auf nil setzen }
  InfoHeader:= nil;
end;

procedure TRCIBitmap.LoadBMPFromFile(const Filename: string);
var
  f         : FILE;         { Ladefile }
  a         : BYTE;         { simpler Laufindex }
  InfoHeader: ^TBitmapInfo; { InfoHeader mit RedColorTable }
begin
  { temporren Bitmapinfo-Header erzeugen }
  GetMem(InfoHeader, SizeOf(TBitmapInfo)+ 1020);
  { Ressourcen checken }
  CheckMem;
  { Datei ffnen }
  AssignFile(f, Filename);
  Reset(f, 1);
  { DIB-Header lesen }
  BlockRead(f, DIBHeader, SizeOf(DIBHeader));
  { DIBInfoHeader lesen }
  BlockRead(f, DIBInfoHeader^, SizeOf(TBitmapInfo)- 4);
  { temporren BitmapInfoHeader mit gleichen Daten belegen }
  InfoHeader^:= DIBInfoHeader^;
  { Colortable lesen }
  BlockRead(f, DIBInfoHeader^.bmiColors, 1024);
  for a:= 0 to 255 do
    { interne Palette mit Daten fllen }
    InfoHeader^.bmiColors[a]:= RedColorTable[a];
  { Originalbitmap in OrigDIBBits speichern }
  with DIBInfoHeader.bmiHeader do
  begin
      { ist Breite durch vier teilbar? }
      if  biWidth mod 4 <> 0 then
        { ... Nein! dann berlauf ausrechnen }
        MehrPix:= 4- (biWidth mod 4)
      else
        { ... ansonsten Pixelberlauf auf null setzen }
        MehrPix:= 0;
    { Gre fr DIB reservieren }
    GetMem(OrigDIBBits, (biWidth+ MehrPix)* biHeight);
    { DIB-Daten lesen }
    BlockRead(f, OrigDIBBits^, (biWidth+ MehrPix)* biHeight);
  end;
  { Datei schlieen }
  CloseFile(f);
  { DIB-Handle erzeugen }
  with DIBInfoHeader^.bmiHeader do
  begin
    { Bitmap erzeugen }
    DIB:= CreateDIBSection(Canvas.Handle, InfoHeader^,
                           DIB_RGB_COLORS, NewDIBBits, nil, 0);
    { Bitmap-Bits setzen }
    SetDIBits(Canvas.Handle, DIB, 0, biHeight, OrigDIBBits,
              InfoHeader^, DIB_RGB_COLORS);
  end;
  { Objekt selektieren }
  SelectObject(Canvas.Handle, DIB);
  with DIBInfoHeader^.bmiHeader do
    { Gre fr DIB im  RAM freigeben }
    FreeMem(OrigDIBBits, (biWidth+ MehrPix)* biHeight);
  { Bitmapdaten existieren nicht mehr }
  OrigDIBBits:= nil;
  { temporren Bitmapinfo-Header wieder lschen }
  FreeMem(InfoHeader, SizeOf(TBitmapInfo)+ 1020);
  { InfoHeaderangaben resetten }
  InfoHeader:= nil;
  { letztes Format war BMP }
  LastFormat:= BMPFormat;
end;

function TRCIBitmap.LoadPCXFromFile(const Filename: string): BOOLEAN;
var
  f         : FILE;         { Ladefile }
  a         : BYTE;         { simpler Laufindex }
  Color     : BYTE;         { Letztes gelesenes Byte }
  count     : Word;         { So oft wird Farbbyte geschrieben }
  OldPos    : Longint;      { alte Pos. in Datei }
  InfoHeader: ^TBitmapInfo; { InfoHeader mit RedColorTable }
  PCXInfo   : PCXHeader;    { PCX-InfoHeader }
  DrawTo    : PChar;        { Pointer auf Daten }
  SavedLine : POINTER;      { alte Zeilen zum Tausch }
  UsedMem   : Longint;      { Soviel Speciher wird fr PCX bentigt }
  x, y      : Longint;      { aktuelle x-, y-Position }
  Breite    : Longint;      { Breite des aktuellen PCX-Files }
  Hoehe     : Longint;      { Hhe des aktuellen PCX-Files }
  Div2      : BOOLEAN;      { war Breite des PCX durch zwei teilbar? }
begin
  { kein Fehler beim laden }
  LoadPCXFromFile:= TRUE;
  { Ressourcen checken }
  CheckMem;
  { temporren Bitmapinfo-Header erzeugen }
  GetMem(InfoHeader, SizeOf(TBitmapInfo)+ 1020);
  { Datei ffnen }
  AssignFile(f, Filename);
  Reset(f, 1);
  { PCX-Dateiinfo einholen }
  BlockRead(f, PCXInfo, SizeOf(PCXInfo));
  { mehr bzw. weniger als 256 Farbenpcx? }
  if (PCXInfo.Planes = 1) and
     (PCXInfo.BitProPixel = 8) then
  begin
    { ... wie ist Farbtabelle abgespeichert? }
    if PCXInfo.BitProPixel = 8 then
    begin
      { alte Position speichern }
      OldPos:= System.FilePos(f);
      { ... und Positionszeiger setzen }
      System.Seek(f, FileSize(f)- 769);
      { Kennungsbyte fr Palette lesen }
      BlockRead(f, Color, 1);
      { war das berhaupt eine Kennung? }
      if Color = 12 then
      begin
        { Colortable lesen }
        for a:= 0 to 255 do
        begin
          { Rotwert lesen }
          BlockRead(f, DIBInfoHeader^.bmiColors[a].rgbRed, 1);
          { Grnwert lesen }
          BlockRead(f, DIBInfoHeader^.bmiColors[a].rgbGreen, 1);
          { Blauwert lesen }
          BlockRead(f, DIBInfoHeader^.bmiColors[a].rgbBlue, 1);
        end;
      end
      else
        { Colortable belegen }
        for a:= 0 to 15 do
        begin
          { Rotwert belegen }
          DIBInfoHeader^.bmiColors[a].rgbRed:=
            PCXInfo.Farbe[a].rgbRed;
          { Grnwert belegen }
          DIBInfoHeader^.bmiColors[a].rgbGreen:=
            PCXInfo.Farbe[a].rgbGreen;
          { Blauwert belegen }
          DIBInfoHeader^.bmiColors[a].rgbBlue:=
            PCXInfo.Farbe[a].rgbBlue;
        end;
      { ... und Positionszeiger zurcksetzen }
      System.Seek(f, OldPos);
    end
    else
      { Colortable belegen }
      for a:= 0 to 15 do
      begin
        { Rotwert belegen }
        DIBInfoHeader^.bmiColors[a].rgbRed:=
          PCXInfo.Farbe[a].rgbRed;
        { Grnwert belegen }
        DIBInfoHeader^.bmiColors[a].rgbGreen:=
          PCXInfo.Farbe[a].rgbGreen;
        { Blauwert belegen }
        DIBInfoHeader^.bmiColors[a].rgbBlue:=
          PCXInfo.Farbe[a].rgbBlue;
      end;
    { Breite / Hhe berechnen }
    Breite:= PCXInfo.xMax- PCXInfo.xMin+ 1;
    { ist Breite durch vier teilbar? }
    if  Breite mod 2 <> 0 then
    begin
      { ... Nein! dann berlauf ausrechnen }
      Breite:= Breite+ 1;
      { Breite war nicht durch zwei teilbar! }
      Div2:= FALSE;
    end
    else
      { ... war durch zwei teilbar! }
      Div2:= TRUE;
    Hoehe:= PCXInfo.yMax- PCXInfo.yMin+ 1;
    { ist Breite durch vier teilbar? }
    if  Breite mod 4 <> 0 then
      { ... Nein! dann berlauf ausrechnen }
      MehrPix:= 4- (Breite mod 4)
    else
      { ... ansonsten Pixelberlauf auf null setzen }
      MehrPix:= 0;
    { DIB-Header mit Werten belegen }
    FillHeader(Breite+ MehrPix, Hoehe);
    { InfoHeader kopieren }
    InfoHeader^:= DIBInfoHeader^;
    { Interne Farbpalette setzen }
    for a:= 0 to 255 do
      { interne Palette mit Daten fllen }
      InfoHeader^.bmiColors[a]:= RedColorTable[a];
    { Originalbitmap in OrigDIBBits speichern }
    with DIBInfoHeader.bmiHeader do
    begin
      { bentigten Speicher berechnen }
      UsedMem:= biWidth* biHeight;
      { Speicherbitmap erzeugen }
      GetMem(OrigDIBBits, UsedMem);
      { Zeiger zuweisen }
      DrawTo:= OrigDIBBits;
      { aktuelle Position null setzen }
      x:= 0;
      y:= 0;
      { Schleife bis zum Ende durchlaufen }
      repeat
        { Farbbyte lesen }
        BlockRead(f, Color, 1);
        { Ist countflag gesetzt }
        if (Color > 192) and (PCXInfo.Encod <> 0) then
        begin
          { Farbe wird sooft gezeichnet }
          count:= Color- 192;
          { Farbbyte lesen }
          BlockRead(f, Color, 1);
        end
        else
          count:= 1;
        { mit Farbbytes fllen }
        FillChar((DrawTo+ y* biWidth+ x)^, count, Color);
        { x-Position erhhen }
        x:= x+ count;
        { aktuelle Position erhhen }
        if x >= Breite then
        begin
          x:= 0;
          y:= y+ 1;
        end;
      until y > Hoehe;
      { Speicher fr Zeilentausch reservieren }
      GetMem(SavedLine, biWidth);
      { nur die hlfte der Zeilen durchlaufen und   }
      { jeweils die obere gegen die untere tauschen }
      for y:= 0 to biHeight div 2- 1 do
      begin
        { Linie oben speichern }
        Move((DrawTo+ y* biWidth)^, SavedLine^, biWidth);
        { ... und untere Linie gegen obere tauschen }
        Move((DrawTo+ UsedMem- (y+ 1)* biWidth)^,
             (DrawTo+ y* biWidth)^, biWidth);
        { ... und obere Linie gegen untere tauschen }
        Move(SavedLine^, (DrawTo+ UsedMem- (y+ 1)*
             biWidth)^, biWidth);
      end;
      { Speicher fr Zeilentausch freigeben }
      FreeMem(SavedLine, biWidth);
      { Linie nicht mehr im Speicher vorhanden }
      SavedLine:= nil;
      { Anzahl zustzlicher Pixel wieder abziehen }
      biWidth:= biWidth- MehrPix;
      { war Breite durch zwei teilbar? }
      if Div2 = FALSE then
        { ... dann noch eins abziehen! }
        biWidth:= biWidth- 1;
    end;
    { Datei schlieen }
    CloseFile(f);
    { DIB-Handle erzeugen }
    with DIBInfoHeader^.bmiHeader do
    begin
      { Bitmap erzeugen }
      DIB:= CreateDIBSection(Canvas.Handle, InfoHeader^,
                             DIB_RGB_COLORS, NewDIBBits, nil, 0);
      { Bitmap-Bits setzen }
      SetDIBits(Canvas.Handle, DIB, 0, biHeight, OrigDIBBits,
                InfoHeader^, DIB_RGB_COLORS);
    end;
    { Objekt selektieren }
    SelectObject(Canvas.Handle, DIB);
    with DIBInfoHeader^.bmiHeader do
      { Gre fr DIB im  RAM freigeben }
      FreeMem(OrigDIBBits, (biWidth+ MehrPix)* biHeight);
    { Bitmapdaten gibt es nicht mehr }
    OrigDIBBits:= nil;
    { temporren Bitmapinfo-Header wieder lschen }
    FreeMem(InfoHeader, SizeOf(TBitmapInfo)+ 1020);
    { kein InfoHeader mehr vorhanden }
    InfoHeader:= nil;
    { letztes Format war PCX }
    LastFormat:= PCXFormat;
  end
  { mehr bzw. weniger als 256 Farbenpcx? }
  else
  begin
    { Datei schlieen }
    CloseFile(f);
    { Fehlermeldung }
    Application.MessageBox('This isn`t a PC Paintbrush file '+
                           'with 256 colors!', 'Error', MB_OK);
    { kein Fehler beim laden }
    LoadPCXFromFile:= FALSE;
  end;
end;

function TRCIBitmap.LoadFromFile(const Filename: string): boolean;
var
  f     : FILE; { Ladefile }
  Format: Word; { wird zur Format erkennung bentigt }
begin
  { Wartenmauszeiger setzen }
  Screen.Cursor:= crHourglass;
  { kein Fehler beim laden aufgetreten }
  LoadFromFile:= TRUE;
  { existiert zu ladende Datei? }
  if FileExists(Filename) = TRUE then
  begin
    { Standard: Format unbekannt }
    Format:= 0;
    { Datei zur Formatprfung ffnen }
    { ... dazu Handle zuweisen }
    AssignFile(f, Filename);
    { ... und ffnen }
    Reset(f, 1);
    { ... die ersten 2 Byte zur Erkennung lesen }
    BlockRead(f, Format, SizeOf(Format));
    { ... wieder schlieen }
    CloseFile(f);
    { Format auf Bitmap testen (19778 = 'BM') }
    if Format = 19778 then
      { ... dann als Bitmap laden }
      LoadBMPFromFile(Filename)
    else
      { ... auf PCX-Datei testen (niederwertiges Byte = $A) }
      if Lo(Format) = $A then
        { ... dann als PCX laden }
        LoadFromFile:= LoadPCXFromFile(Filename)
      else
      begin
        { ... ansonsten Fehlermeldung (Unbekanntes Format) }
        Application.MessageBox('Unknown fileformat!', 'Error', MB_OK);
        { Fehler beim laden aufgetreten }
        LoadFromFile:= FALSE;
      end;
  end
  else
  begin
    { ... ansonsten Fehlermeldung anzeigen }
    Application.MessageBox('File not found!', 'Error', MB_OK);
    { Fehler beim laden aufgetreten }
    LoadFromFile:= FALSE;
  end;
  { Standardmauszeiger setzen }
  Screen.Cursor:= crDefault;
end;

procedure TRCIBitmap.SaveBMPToFile(const Filename: string);
var
  f: FILE; { Savefile }
  a: BYTE; { simpler Laufindex }
begin
  { neue Datei erzeugen }
  AssignFile(f, Filename);
  ReWrite(f, 1);
  { DIB-Header schreiben }
  BlockWrite(f, DIBHeader, SizeOf(TBitmapFileHeader));
  { DIB-Infoheader+ Colortable schreiben }
  BlockWrite(f, DIBInfoHeader^, SizeOf(TBitmapInfo)+ 1020);
  with DIBInfoHeader.bmiHeader do
  begin
    { ist Breite durch vier teilbar? }
    if biWidth mod 4 <> 0 then
      { ... Nein! dann berlauf ausrechnen }
      MehrPix:= 4- (biWidth mod 4)
    else
      { ... ansonsten Pixelberlauf auf null setzen }
      MehrPix:= 0;
    { DIB-Daten schreiben }
    BlockWrite(f, NewDIBBits^, (biWidth+ MehrPix)* biHeight);
  end;
  { Datei schlieen }
  CloseFile(f);
  { letztes Format war BMP }
  LastFormat:= BMPFormat;
end;

procedure TRCIBitmap.SavePCXToFile(const Filename: string);
var
  f      : FILE;      { Savefile }
  a      : BYTE;      { simpler Laufindex }
  Header : PCXHeader; { PCX-Header }
  AktByte: Longint;   { aktuelles Byte }
  x, y   : Longint;   { Spalte in aktueller Zeile }
  ColTab : array[0..768] of BYTE;
                      { Farbpalette }
  Data   : PChar;     { Pointer auf Daten }
  Color  : BYTE;      { aktuelle Farbe }
  Counter: BYTE;      { Counter }
  LastCol: BYTE;      { letzte Farbe }
  Breite : Longint;   { Breite des PCX-Pictures }
begin
  { neue Datei erzeugen }
  AssignFile(f, Filename);
  ReWrite(f, 1);
  { PCX-Header fllen }
  with Header, DIBInfoHeader^.bmiHeader do
  begin
    Id:= 10;            { PCX-Kennung }
    Ver:= 5;            { Versionskennung }
    Encod:= 1;          { Runlength-Encoding }
    BitProPixel:= 8;    { 256-Farben PCX }
    xMin:= 0;           { BS-Koordinaten }
    yMin:= 0;           { BS-Koordinaten }
    xMax:= biWidth- 1;  { BS-Koordinaten }
    yMax:= biHeight- 1; { BS-Koordinaten }
    hres:= biWidth;     { Horizontale Auflsung }
    vres:= biHeight;    { Vertikale Auflsung }
    { in Farbpalette ersten 16 Eintrge speichern }
    for a:= 0 to 15 do
    begin
      Farbe[a].rgbRed:= DIBInfoHeader^.bmiColors[a].rgbRed;
      Farbe[a].rgbGreen:= DIBInfoHeader^.bmiColors[a].rgbGreen;
      Farbe[a].rgbBlue:= DIBInfoHeader^.bmiColors[a].rgbBlue;
    end;
    Fill:= 0;           { Reserviertbyte setzen }
    Planes:= 1;         { Anzahl Planes }
    ByteProZeile:= biWidth;
                        { Anzahl Bytes pro Zeile }
    PalInfo:= 1;        { Farbpalette }
    { Reserviertbytes setzen }
    for a:= 0 to 57 do
      Fill3[a]:= 0;
  end;
  { DIB-Header schreiben }
  BlockWrite(f, Header, SizeOf(PCXHeader));
  { Breite berechnen }
  Breite:= Header.xMax- Header.xMin+ 1;
  { ist diese Breite durch zwei teilbar?? }
  if Breite mod 2 <> 0 then
    { ... Nein!! ... auf nchsten Teiler aufrunden }
    Breite:= Breite+ 1;
  { Datenpointer setzen }
  Data:= NewDIBBits;
  with DIBInfoHeader.bmiHeader do
  begin
    { ist Breite des Bitmaps durch vier teilbar? }
    if biWidth mod 4 <> 0 then
      { ... Nein! dann berlauf ausrechnen }
      MehrPix:= 4- (biWidth mod 4)
    else
      { ... ansonsten Pixelberlauf auf null setzen }
      MehrPix:= 0;
    { jede Zeile durchlaufen }
    for y:= biHeight- 1 downto 0 do
    begin
      { aktuelle Adresse der aktuellen Zeile berechnen }
      AktByte:= y* (biWidth+ MehrPix);
      { erstes Farbbyte holen }
      Move((Data+ AktByte)^, Color, 1);
      { letzte Farbe auf aktuelle Farbe speichern }
      LastCol:= Color;
      { Zhler auf ins setzen }
      Counter:= 1;
      { ... und jetzt Spaltendurchlauf }
      for x:= 1 to Breite do
      begin
        { nchstes Farbbyte holen }
        Move((Data+ AktByte+ x)^, Color, 1);
        { unterschiedliche Farben }
        if (LastCol <> Color) or (Counter >= 63) or
           (x = Breite) then
        begin
          { nur Counter abspeichern, wenn     }
          { mehr als ein mal und Farbe >= 192 }
          if (Counter > 1) or (LastCol >= 192) then
          begin
            { ... dann Counter speichern }
            Counter:= Counter+ 192;
            BlockWrite(f, Counter, 1);
          end;
          { ... und Farbe speichern }
          BlockWrite(f, LastCol, 1);
          { Zhler auf 0 zurcksetzen }
          Counter:= 0;
        end;
        { Zhler erhhen }
        INC(Counter);
        { letzte Farbe speichern }
        LastCol:= Color;
      end;
    end;
  end;
  { Farbpalette belegen }
  { Kennung fr Farbpalette }
  ColTab[0]:= 12;
  { ... belegen }
  for a:= 0 to 255 do
  begin
    ColTab[a* 3+ 1]:= DIBInfoHeader^.bmiColors[a].rgbRed;
    ColTab[a* 3+ 2]:= DIBInfoHeader^.bmiColors[a].rgbGreen;
    ColTab[a* 3+ 3]:= DIBInfoHeader^.bmiColors[a].rgbBlue;
  end;
  { Colortable schreiben }
  BlockWrite(f, ColTab, SizeOf(ColTab));
  { Datei schlieen }
  CloseFile(f);
  { letztes Format war PCX }
  LastFormat:= PCXFormat;
end;

function TRCIBitmap.SaveToFile(const Filename: string;
                               Format: Integer): Boolean;
begin
  { Wartenmauszeiger setzen }
  Screen.Cursor:= crHourglass;
  { Datei konnte korrekt gespeichert werden }
  SaveToFile:= TRUE;
  { Bild als BMP abspeichern? }
  if Format = BMPFormat then
    { ... dann speichern }
    SaveBMPToFile(Filename)
  else
    { oder Bild als PCX abspeichern? }
    if Format = PCXFormat then
      { ... dann speichern }
      SavePCXToFile(Filename)
    else
      { oder sogar automatisches Formatspeichern? }
      if Format >= AutoDetectFormat then
      begin
        { ... dann auf Endung BMP berprfen }
        if (UpperCase(ExtractFileExt(Filename)) = '.BMP') or
           (UpperCase(ExtractFileExt(Filename)) = '.DIB') then
          { ... dann als BMP speichern }
          SaveBMPToFile(Filename)
        else
          { ... dann auf Endung PCX berprfen }
          if UpperCase(ExtractFileExt(Filename)) = '.PCX' then
            { ... dann als PCX speichern }
            SavePCXToFile(Filename)
          else
          begin
            { ... ansonsten Fehlermeldung (Unbekanntes Format) }
            Application.MessageBox('Unknown fileformat! Couldnt '+
                                   'save file!', 'Error', MB_OK);
            { Dateifehler beim speichern unterlaufen }
            SaveToFile:= FALSE;
          end;
      end;
  { Standardmauszeiger setzen }
  Screen.Cursor:= crDefault;
end;

procedure TRCIBitmap.CopyToClipboard(Pos: TRect);
var
  ToPoint   : POINTER; { Position des Handles als Pointer }
  NewPic    : TRCIBitmap;
                       { neuer InfoHeader inkl. Bitmapdaten }
  AktPal    : TRCIEntries;
                       { aktuelle Farbpalette }
  CopyHandle: THandle; { Handle, zu dem Daten kopiert werden }
  NeededSize: Longint; { Dieser Speicher wird reserviert }
  Width     : Longint; { Breite des zu kopierenden Ausschnittes }
  Height    : Longint; { Hhe des zu kopierenden Ausschnittes }
  WidthBMP  : Longint; { Breite des Bitmaps }
                       { Win95 besitzt die Eigenart, Bilder die }
                       { nicht aus einer Breite bestehen, die   }
                       { durch 4 teilbar ist fehl zu inter-     }
                       { pretieren.                             }
begin
  { Wartenmauszeiger setzen }
  Screen.Cursor:= crHourglass;
  { neues Bild erzeugen }
  NewPic:= TRCIBitmap.Create;
  { aktuelle Farbwerte bernehmen }
  for a:= 0 to 255 do
  begin
    AktPal[a].rgbRed:= DIBInfoHeader^.bmiColors[a].rgbRed;
    AktPal[a].rgbGreen:= DIBInfoHeader^.bmiColors[a].rgbGreen;
    AktPal[a].rgbBlue:= DIBInfoHeader^.bmiColors[a].rgbBlue;
  end;
  { Breite berechnen }
  Width:= Pos.Right- Pos.Left+ 1;
  { ist akt. Breite durch 4 teilbar? }
  if (Width mod 4) <> 0 then
    { ... nein, dann fehlenden Rest dazuaddieren }
    WidthBMP:= Width+ 4- (Width mod 4)
  else
    { ... ansonsten Breite bernehmen }
    WidthBMP:= Width;
  { Hhe berechnen }
  Height:= Pos.Bottom- Pos.Top+ 1;
  { neues Bild erzeugen }
  NewPic.NewPicture(WidthBMP, Height, 247, AktPal);
  { Farbreserviertbytes setzen (Kennung fr mein Prog) }
  for a:= 0 to 255 do
    { ... setzen }
    NewPic.DIBInfoHeader^.bmiColors[a].rgbReserved:=
      ClipboardDetect;
  with NewPic.DIBInfoHeader^.bmiHeader do
  begin
    { richtige Breite setzen }
    biWidth:= Width;
{ ------------------- Datenkopierroutine -------------------- }
    { ... und Daten kopieren }
    BitBlt(NewPic.Canvas.Handle, 0, 0,
           Width, Height, Canvas.Handle,
           Pos.Left, Pos.Top, SRCCOPY);
    { bentigten RAM-Speicher errechnen }
    NeededSize:= SizeOf(TBitmapInfo)+
                   WidthBMP* Height+ 1020;

{ ---------------- Daten in Stream kopieren ----------------- }
    { Clipboarddata erzeugen }
    ClipboardData:= TMemoryStream.Create;
    { Gre setzen }
    ClipboardData.SetSize(NeededSize);
    { Daten kopieren 1 (InfoHeader+ Colors) }
    ClipboardData.WriteBuffer(NewPic.DIBInfoHeader^,
                              SizeOf(TBitmapInfo)+ 1020);
    { Daten kopieren 2 (Datenbits) }
    ClipboardData.WriteBuffer(NewPic.NewDIBBits^,
                              WidthBMP* Height);

{ ---------------- Daten in RAM-Handle kopieren ------------- }
    { Block allokieren }
    CopyHandle:= GlobalAlloc(GMEM_MOVEABLE, NeededSize);
    { konnte Pointer belegt werden? }
    if CopyHandle <> 0 then
    begin
      { ... Handle in Pointer konvertieren }
      ToPoint:= GlobalLock(CopyHandle);
      { Daten kopieren }
      Move(ClipboardData.Memory^, ToPoint^, NeededSize);

      { Clipboard ffnen }
      if OpenClipboard(0) = TRUE then
      begin
        { Clipboarddaten setzen }
        SetClipboardData(CF_DIB, CopyHandle);
        { Clipboard schlieen }
        CloseClipboard;
      end
      else
        { ... ansonsten Fehlermeldung }
        Application.MessageBox('Couldnt open clipboard',
                               'Error', MB_OK);
      { ... Pointer in Handle konvertieren }
      GlobalUnlock(CopyHandle);
      { Block reallokieren }
      CopyHandle:= GlobalFree(CopyHandle);
    end
    else
      { ... ansonsten Fehlermeldung }
      Application.MessageBox('Couldnt allocate memory for DIB',
                             'Error', MB_OK);
    { Clipboarddata lschen }
    ClipboardData.Destroy;
  end;
  { Bild im RAM wieder lschen }
  NewPic.Destroy;
  { Standardmauszeiger setzen }
  Screen.Cursor:= crDefault;
end;

{ Gibt zurck, welches Programm Zwischenablage gefllt hat }
{ (TRUE = eigenes [TRCIBitmap]; FALSE = ein anderes)       }
function TRCIBitmap.PastePicture: BOOLEAN;
var
  FromPPos  : PChar;   { Position des Pointers als Longint (Quelle) }
  ToPPos    : PChar;   { Position des Pointers als Longint (Ziel) }
  y         : Longint; { simpler Laufindex }
  FromPoint : POINTER; { von hier wird kopiert }
  CopyHandle: THandle; { Handle, zu dem Daten kopiert werden }
  BMPHeader : ^TBitmapInfo;
                       { gelesener Infoheader }
  Width     : Longint; { Breite des zu kopierenden Ausschnittes }
  WidthBMP  : Longint; { Breite des Bitmaps }
  Height    : Longint; { Hhe des zu kopierenden Ausschnittes }
  AktPal    : TRCIEntries;
                       { aktuelle Farbpalette }
begin
  { Clipboard ffnen }
  if OpenClipboard(0) = TRUE then
  begin
    { berprfen, ob Format in Clipboard }
    if IsClipboardFormatAvailable(CF_DIB) = TRUE then
    begin
      { Ja!! Dann einfgen }
      CopyHandle:= GetClipboardData(CF_DIB);
      { ... Handle in Pointer konvertieren }
      FromPoint:= GlobalLock(CopyHandle);
      { Speicher fr Header reservieren }
      GetMem(BMPHeader, SizeOf(TBitmapInfo)+ 1020);
      { Header kopieren }
      Move(FromPoint^, BMPHeader^,
           SizeOf(TBitmapInfo)+ 1020);
      { Breite berechnen }
      Width:= BMPHeader^.bmiHeader.biWidth;
      { Hhe berechnen }
      Height:= BMPHeader^.bmiHeader.biHeight;
      { aktuelle Farbwerte bernehmen }
      for a:= 0 to 255 do
      begin
        AktPal[a].rgbRed:= BMPHeader^.bmiColors[a].rgbRed;
        AktPal[a].rgbGreen:= BMPHeader^.bmiColors[a].rgbGreen;
        AktPal[a].rgbBlue:= BMPHeader^.bmiColors[a].rgbBlue;
      end;
      { ist akt. Breite durch 4 teilbar? }
      if (Width mod 4) <> 0 then
        { ... nein, dann fehlenden Rest dazuaddieren }
        WidthBMP:= Width+ 4- (Width mod 4)
      else
        { ... ansonsten Breite bernehmen }
        WidthBMP:= Width;
      { neues Bild erzeugen }
      NewPicture(Width, Height, 247, AktPAL);
      { Pointer als RAMPos bestimmen }
      FromPPos:= FromPoint;
      ToPPos:= NewDIBBits;
      { jede Zeile in Bitmap kopieren }
      for y:= 0 to Height- 1 do
        { Zeile x kopieren }
        Move((FromPPos+ y* WidthBMP+ SizeOf(TBitmapInfo)+ 1020)^,
             (ToPPos+ y* WidthBMP)^, Width);
      { berprfen, woher Bitmap kommt }
      { Annahme: von mir }
      PastePicture:= TRUE;
      { Annahme berprfen }
      for a:= 0 to 255 do
        { Fehler gefunden --> nicht von mir }
        if BMPHeader^.bmiColors[a].rgbReserved <> ClipboardDetect then
        begin
          { nicht von mir!! }
          PastePicture:= FALSE;
          { Schleife verlassen }
          Break;
        end;
      { Speicher fr Header wieder freigeben }
      FreeMem(BMPHeader, SizeOf(TBitmapInfo)+ 1020);
      { kein BMP-Header mehr vorhanden }
      BMPHeader:= nil;
      { ... Pointer in Handle konvertieren }
      GlobalUnlock(CopyHandle);
    end
    else
      { ... ansonsten Fehlermeldung }
      Application.MessageBox('Couldnt found a DIB in clipboard',
                             'Error', MB_OK);
    { Clipboard wieder schlieen }
    CloseClipboard;
  end
  else
    { ... ansonsten Fehlermeldung }
    Application.MessageBox('Couldnt open clipboard',
                           'Error', MB_OK);
end;

begin
  { Vorbelegung von RedColorTable }
  for a:= 0 to 255 do
  begin
    { Rot-Anteil (eigentlich spter Farbnummer) }
    RedColorTable[a].rgbRed:= a;
    { Grn-Anteil (immer 0 setzen!) }
    RedColorTable[a].rgbGreen:= 0;
    { Blau-Anteil (immer 0 setzen!) }
    RedColorTable[a].rgbBlue:= 0;
    { Reserviert-Byte (immer 0 setzen!) }
    RedColorTable[a].rgbReserved:= 0;
  end;
end.

